import os
import zoneinfo


# For logging
STREAMING_LOG_LEVEL = int(os.getenv("STREAMING_LOG_LEVEL", "30"))
FILE_LOG_LEVEL = int(os.getenv("FILE_LOG_LEVEL", "10"))
LOG_FILEPATH = os.getenv("LOG_FILEPATH")
LOG_FMT = os.getenv("LOG_FMT", "%(asctime)s [%(name)s | %(levelname)s]: %(message)s")
LOG_DATEFMT = os.getenv("LOG_DATEFMT", "%Y-%m-%dT%H:%M:%SZ")
TZ = zoneinfo.ZoneInfo(os.getenv("TZ") or "UTC")
ENABLE_PRERELEASE_WARNING = os.getenv("ENABLE_PRERELEASE_WARNING", "true").lower() == "true"



from typing import Union

import numpy as np
import pandas as pd
from tenacity import retry, stop_after_attempt

from archeo.core.forward.resampler import ISData
from archeo.utils.helper import pre_release
from archeo.utils.parallel import multithread_run


@retry(stop=stop_after_attempt(3))
def _resample_and_get_bayes_factor(
    candidate_prior: pd.DataFrame, prior: pd.DataFrame, posterior: pd.DataFrame
) -> float:
    """Resample the prior and posterior samples and compute the Bayes factor."""

    data = ISData(
        new_prior_samples=candidate_prior.sample(len(candidate_prior), replace=True),
        posterior_samples=posterior.sample(len(posterior), replace=True),
        prior_samples=prior.sample(len(prior), replace=True),
    )
    return data.get_bayes_factor()


def _extract_kick_thresholds(
    kicks: Union[list[pd.Series], pd.Series],
) -> tuple[float, float]:
    """Extract the lower and upper bounds of the kick thresholds."""

    if isinstance(kicks, pd.Series):
        # Here we ensure at least 1000 samples for the kick threshold
        k_lb = kicks.quantile(1000 / len(kicks))
        k_ub = kicks.max()
    else:
        # Here we ensure at least 1000 samples for each kick threshold
        k_lb = max([ks.quantile(1000 / len(ks)) for ks in kicks])
        k_ub = max([ks.max() for ks in kicks])

    return k_lb, k_ub


@pre_release
def get_bayes_factor_over_escape_velocity(
    prior: pd.DataFrame,
    posterior: pd.DataFrame,
    candidate_prior: dict[str, pd.Series],
    kicks: Union[list[pd.Series], pd.Series],
    n_trials: int = 500,
    ref_bayes_factor: float = 1.0,
) -> dict[str, Union[list[float], list[list[float]]]]:
    """Compute the Bayes factor over a range of kick thresholds.

    Args:
        prior (pd.DataFrame): Prior samples.
        posterior (pd.DataFrame): Posterior samples.
        candidate_prior (dict[str, pd.Series]): Candidate prior samples.
        kicks (Union[list[pd.Series], pd.Series]): Kick thresholds.
        n_trials (int): Number of trials to run for resampling.
        ref_bayes_factor (float): Reference Bayes factor to normalize the results.

    Returns:
        dict: A dictionary containing the kick cutoff values, Bayes factor samples,
              median Bayes factor, and the 5th and 95th percentiles of the Bayes factor samples.
    """

    k_lb, k_ub = _extract_kick_thresholds(kicks)
    _kicks = (
        # Sample 100 evenly spaced values in log space between k_lb and k_ub
        np.logspace(np.log10(k_lb), np.log10(k_ub), 100)
        .astype(float)
        .tolist()
    )
    _bf_samples: list[list[float]] = []

    for k in _kicks:
        # Mark index of the samples that are below the kick threshold
        if isinstance(kicks, pd.Series):
            idx = kicks < k
            _candidate_prior = {col: ds[idx] for col, ds in candidate_prior.items()}
        else:
            _candidate_prior = {}
            for col, ds in candidate_prior.items():
                idx = kicks[col.endswith("2")] < k
                _candidate_prior[col] = ds[idx]

        _bfs = multithread_run(
            func=_resample_and_get_bayes_factor,
            input_kwargs=[
                {
                    "candidate_prior": pd.DataFrame(_candidate_prior),
                    "prior": prior,
                    "posterior": posterior,
                }
                for _ in range(n_trials)
            ],
        )

        _bfs = [_bf / ref_bayes_factor for _bf in _bfs]
        _bf_samples.append(_bfs)

    return {
        "kick_cutoff": _kicks,
        "bayes_factor_samples": _bf_samples,
        "bayes_factor_median": [np.median(bfs) for bfs in _bf_samples],
        "bayes_factor_low": [np.percentile(bfs, q=5) for bfs in _bf_samples],
        "bayes_factor_high": [np.percentile(bfs, q=95) for bfs in _bf_samples],
    }



from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.colors import LinearSegmentedColormap

from archeo.constants import Columns as C
from archeo.constants import Suffixes as S
from archeo.schema import Labels
from archeo.visualization import base


# Color map for the scatter plot
WHITE_VIRIDIS = LinearSegmentedColormap.from_list(
    "white_viridis",
    [  # type: ignore
        (0, "#ffffff"),
        (1e-20, "#440053"),
        (0.2, "#404388"),
        (0.4, "#2a788e"),
        (0.6, "#21a784"),
        (0.8, "#78d151"),
        (1, "#fde624"),
    ],
    N=256,
)


def distribution_summary(
    df: pd.DataFrame,
    filename="distribution_summary",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the distribution of the black hole parameters.

    Args:
        df (pd.DataFrame): The prior/posterior dataframe.
        filename (str): Filename of the figure.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    labels = Labels("Distribution of black-hole parameters")
    col_to_labels = {
        C.MASS_RATIO: "Parent Mass Ratio $q$",
        S.FINAL(C.MASS): r"Remnant Mass $m_f$ [$M_{\odot}$]",
        S.FINAL(C.KICK): "Recoil Kick $v_f$ [$kms^{-1}$]",
        S.FINAL(C.SPIN_MAG): "Spin $\\chi_f$",
    }
    fig, axes = base.initialize_plot(nrows=4, ncols=1, figsize=(6, 8), labels=labels)
    for idx, (col, xlabel) in enumerate(col_to_labels.items()):
        sns.histplot(df[col], ax=axes[idx], element="step", fill=False, stat="density")
        axes[idx].set(xlabel=xlabel, ylabel="PDF")

    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, axes)


def kick_against_spin_cmap(
    df: pd.DataFrame,
    filename="kick_against_spin_cmap",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the remnant kick against remnant spin in cmap.

    Args:
        df (pd.DataFrame): The prior/posterior dataframe.
        filename (str): Filename of the figure.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    # mpl_scatter_density import is for ax.scatter_density
    # implicitly used in kick_against_spin
    # pylint: disable-next=unused-import, import-outside-toplevel
    import mpl_scatter_density  # type: ignore

    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(1, 1, 1, projection="scatter_density")
    ax.scatter_density(df[S.FINAL(C.SPIN_MAG)], df[S.FINAL(C.KICK)], cmap=WHITE_VIRIDIS)  # type: ignore
    ax.set(
        title="Remnant Kick against Remnant Spin",
        xlabel="Remnant Spin $\\chi_f$",
        ylabel=r"Remnant Kick $v_f$ [$kms^{-1}$]",
    )
    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, ax)


def kick_distribution_on_spin(
    df: pd.DataFrame,
    filename="kick_on_spin",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the distribution of remnant kick on different spin range.

    Args:
        df (pd.DataFrame): The prior/posterior dataframe.
        filename (str): Filename of the figure.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    labels = Labels(
        title="Remnant Kick Distribution on Different Spin Range",
        xlabel="Remnant Kick $v_f$ [$km/s$]",
        ylabel="PDF",
    )
    fig, ax = base.initialize_plot(figsize=(9, 4), labels=labels)
    bounds = zip(np.linspace(0, 0.9, 10), np.linspace(0.1, 1, 10))

    for low, high in bounds:
        data = df.loc[(low < df[S.FINAL(C.SPIN_MAG)]) & (df[S.FINAL(C.SPIN_MAG)] < high)][S.FINAL(C.KICK)]
        # To avoid extreme density values
        if len(data.index) > 100:
            density, bins = np.histogram(a=data, bins=70, density=True)
            label = rf"$\chi_f$ $\in$ $[{low:.2f}, {high:.2f}]$"
            ax.stairs(density, bins, label=label)

    plt.legend()
    base.clear_default_labels(ax)
    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, ax)



import click

import archeo
from archeo.preset import PRIOR_STORE, get_prior_config


@click.group()
def cli():
    """Command line interface for archeo"""


@cli.command()
@click.option(
    "-n",
    "--name",
    default="default",
    help=f"Preset prior name, available values are {', '.join(PRIOR_STORE)}.",
)
@click.option(
    "-o",
    "--output-dir",
    default=".",
    help="Directory to save the generated prior configuration.",
)
def generate_preset_prior(name: str, output_dir: str):
    """Generate a preset prior."""

    click.echo(f"Generating preset prior: {name}")

    prior_config = get_prior_config(name)
    click.echo(f"Prior configuration: {prior_config}")

    output_path = f"{output_dir}/{name}_prior_config.yaml"
    prior_config.to_yaml(output_path)

    prior = archeo.Prior.from_config(prior_config)

    try:
        prior.to_parquet(f"{output_dir}/{name}_prior.parquet")
        click.echo(f"Prior saved as parquet: {output_dir}/{name}_prior.parquet")
    except ImportError:
        click.echo("Failed to save prior as parquet, please install pyarrow if you want to use this feature.")
        prior.to_csv(f"{output_dir}/{name}_prior.csv")
        click.echo(f"Prior saved as CSV: {output_dir}/{name}_prior.csv")


@cli.command()
@click.option(
    "-f",
    "--filepath",
    required=True,
    help="Path to the prior data.",
)
@click.option(
    "-o",
    "--output-dir",
    default=".",
    help="Directory to save the visualization output.",
)
def visualize_prior(filepath: str, output_dir: str):
    """Visualize the prior distribution."""

    prior = archeo.Prior.from_filepath(filepath)
    archeo.visualize_prior_distribution(prior, output_dir=output_dir)


if __name__ == "__main__":
    cli()



from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.animation import FuncAnimation

from archeo.constants import Columns as C
from archeo.constants import Suffixes as S
from archeo.logger import get_logger


logger = get_logger(__name__)


def animate_remnant_property_change_over_kick(
    df: pd.DataFrame,
    col_name: str,
    kick_lb=50.0,
    kick_width=50.0,
    output_dir: Optional[str] = None,
):
    """Create an animation of the remnant mass/spin distribution change over different kick values

    Args:
        df (pd.DataFrame): DataFrame containing the prior data.
        parameter (str): The remnant property to animate, such as C.MASS.
        kick_lb (float): Lower bound of the kick velocity to consider for the animation.
        kick_width (float): Width of the kick velocity range to consider for the animation.
        output_dir (Optional[str]): Directory to save the animation. If None, the animation will
            not be saved to a file.

    Returns:
        ani (FuncAnimation): The animation object.
    """

    _labels = {
        S.FINAL(C.MASS): "Remnant Mass",
        S.FINAL(C.SPIN_MAG): "Remnant Spin",
        C.MASS_RATIO: "Mass Ratio",
        S.EFF(C.SPIN): "Effective Spin",
        S.PREC(C.SPIN): "Precession Spin",
    }

    if col_name not in _labels:
        logger.warning("Parameter %s is not supported for animation. ", col_name)
        return None

    label = _labels[col_name]

    kick_ub = df[S.FINAL(C.KICK)].max()
    if kick_lb >= kick_ub:
        logger.warning(
            "Kick lower bound %.2f is greater than or equal to upper bound %.2f. " "No animation will be created.",
            kick_lb,
            kick_ub,
        )
        return None

    nbins = int((kick_ub - kick_lb) / kick_width) + 1
    k_bounds = np.linspace(kick_lb, kick_ub, nbins)
    binwidth = (df[col_name].max() - df[col_name].min()) / 70

    fig, ax = plt.subplots(figsize=(8, 6))

    def update(frame):
        ax.clear()
        values = df.loc[df[S.FINAL(C.KICK)] <= k_bounds[frame], col_name]

        if values.nunique() == 1:
            ax.axvline(values.iloc[0], color="red", linestyle="--", linewidth=2)
            ax.set_title(f"{label} (Kick $v_f$ <= {k_bounds[frame]:.2f})")
            ax.set_xlabel(label)
            ax.set_ylabel("Density")
            return ax

        sns.histplot(values, ax=ax, stat="density", binwidth=binwidth, fill=False, element="step")
        ax.set(xlabel=label, ylabel="Density")
        ax.grid(True)
        ax.set_title(f"Distribution of {label} (Kick $v_f$ <= {k_bounds[frame]:.2f})")
        return ax

    ani = FuncAnimation(fig, update, frames=len(k_bounds), interval=100)
    if output_dir:
        filename = label.lower().replace(" ", "_")
        ani.save(f"{output_dir}/{filename}.gif", writer="pillow")

    return ani



import enum
import os
from typing import Union

import pandas as pd

import archeo.logger


local_logger = archeo.logger.get_logger(__name__)

SPEED_OF_LIGHT = 299792.458  # km/s


class Fits(enum.StrEnum):
    """Surrogate models for binary black hole merger simulations.

    Attributes:
        NRSur3dq8Remnant: non precessing BHs with mass ratio<=8, anti-/aligned spin <= 0.8.
        NRSur7dq4Remnant: precessing BHs with mass ratio<=4, generic spin <= 0.8.
        surfinBH7dq2: precessing BHs with mass ratio <= 2, generic spin <= 0.8.

    Details please refer to https://pypi.org/project/surfinBH/.
    """

    NRSUR3DQ8REMNANT = "NRSur3dq8Remnant"
    NRSUR7DQ4REMNANT = "NRSur7dq4Remnant"
    SURFINBH7DQ2 = "surfinBH7dq2"

    def load(self):
        """Load a surfinBH fits.

        Returns:
            fits (surfinBH.surfinBH.SurFinBH): The loaded fits.
        """

        import numpy as np  # pylint: disable=import-outside-toplevel
        import surfinBH  # pylint: disable=import-outside-toplevel

        # TODO: Remove this when SurfinBH upgraded to numpy 2.0
        np.string_ = np.bytes_  # Here we fix the numpy incompatibility issue in SurfinBH

        local_logger.info(
            "Loading surfinBH %s, description: %s.",
            self.value,
            surfinBH.fits_collection[self.value].desc,
        )

        try:
            return surfinBH.LoadFits(self.value)
        except (OSError, KeyError) as e:
            local_logger.error("Failed to load surfinBH %s: %s", self.value, str(e))
            self.clean_up_surfinbh_data()
            return self.load()

    @staticmethod
    def clean_up_surfinbh_data():
        """Clean up the surfinBH data directory.

        We clean up in two situations:
        1. KeyError: this seems to be a bug in surfinBH,
           when installing the latest version (1.2.6).
        2. OSError: this happens when we interrupt the download
           of the surfinBH data files.
        """

        import surfinBH  # pylint: disable=import-outside-toplevel

        # Remove all files in the data directory
        data_dir = f"{os.path.dirname(surfinBH.__file__)}/data"

        if os.path.exists(data_dir):
            for file in os.listdir(data_dir):
                local_logger.warning(
                    "Cleaning up surfinBH data directory: removing %s due to error.",
                    f"{data_dir}/{file}",
                )
                os.remove(f"{data_dir}/{file}")


class EscapeVelocity(enum.Enum):
    """Escape velocity (Unit in km s^-1)"""

    GLOBULAR_CLUSTER = 50.0
    MILKY_WAY = 600.0
    NUCLEAR_STAR_CLUSTER = 1500.0
    ELLIPTICAL_GALAXY = 2500.0

    def label(self) -> str:
        """Return the escape velocity label"""

        if self is EscapeVelocity.GLOBULAR_CLUSTER:
            return "$v_{esc, GC}$"

        if self is EscapeVelocity.MILKY_WAY:
            return "$v_{esc, MW}$"

        if self is EscapeVelocity.NUCLEAR_STAR_CLUSTER:
            return "$v_{esc, NSC}$"

        if self is EscapeVelocity.ELLIPTICAL_GALAXY:
            return "$v_{esc, EG}$"

        raise ValueError(f"Unknown escape velocity {self}")

    def short(self) -> str:
        """Return the short name of the escape velocity"""

        if self is EscapeVelocity.GLOBULAR_CLUSTER:
            return "GC"

        if self is EscapeVelocity.MILKY_WAY:
            return "MW"

        if self is EscapeVelocity.NUCLEAR_STAR_CLUSTER:
            return "NSC"

        if self is EscapeVelocity.ELLIPTICAL_GALAXY:
            return "EG"

        raise ValueError(f"Unknown escape velocity {self}")

    def compute_p2g(self, df: pd.DataFrame) -> float:
        """Return the probability of
        the black hole being a 2nd generation black hole
        under different escape velocity conditions."""

        if df.empty:
            return 0.0

        mask = (
            (df[Suffixes.FINAL(Columns.KICK)] <= self.value)
            & (df[Suffixes.PRIMARY(Columns.MASS)] <= 65)
            & (df[Suffixes.SECONDARY(Columns.MASS)] <= 65)
        )
        return mask.sum() / len(df) * 100.0

    @classmethod
    def to_vlines(cls) -> dict[str, float]:
        """Return a dictionary for vlines plotting

        Returns:
            vlines (Dict[str, float]): The escape velocity vlines.
        """

        return {esc_vel.label(): esc_vel.value for esc_vel in cls}


class Columns(enum.StrEnum):
    """Columns in the prior dataframe"""

    MASS = "m"
    SPIN = "chi"
    SPIN_MAG = "a"
    MASS_RATIO = "q"
    LIKELIHOOD = "l"
    RECOVERY_RATE = "r_rec"
    KS_TEST_FOR_MASS = "ks_test_mf"
    KS_PV_FOR_MASS = "ks_p-value_mf"
    KS_TEST_FOR_SPIN = "ks_test_af"
    KS_PV_FOR_SPIN = "ks_p-value_af"
    SAMPLE_ID = "sample_id"
    KICK = "k"
    VELOCITY = "v"
    MODEL_NAME = "model_name"


class Prefixes(enum.StrEnum):
    """Prefixes for columns"""

    ORIGINAL = "original"
    MIN = "min"

    def __call__(self, column: Union[Columns, str]) -> str:
        """Return the column name with prefix"""

        if isinstance(column, enum.Enum):
            column = column.value

        return f"{self.value}_{column}"


class Suffixes(enum.StrEnum):
    """Suffixes for columns"""

    PRIMARY = "1"
    SECONDARY = "2"
    FINAL = "f"
    EFF = "eff"  # for effective spin
    PREC = "p"  # for precessing spin
    RETAINED = "ret"  # for retained mass
    ESC = "esc"  # for escape velocity

    def __call__(self, column: Union[Columns, str]) -> str:
        """Return the column name with suffix"""

        if isinstance(column, enum.Enum):
            column = column.value

        return f"{column}_{self.value}"



from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd
from scipy.stats import rv_histogram

import archeo.logger
from archeo.schema import Domain
from archeo.utils.helper import pre_release


local_logger = archeo.logger.get_logger(__name__)


def get_histogram(samples: pd.Series, nbins: int, bounds: Domain) -> np.ndarray:
    """Compute the histogram of the samples"""

    hist, edges = np.histogram(samples, bins=nbins, density=True, range=(bounds.low, bounds.high))
    binwidth = edges[1] - edges[0]

    auc = np.sum(hist) * binwidth
    if not np.isclose(auc, 1.0, atol=1e-6):
        msg = f"Invalid probability distribution (AUC={auc:.2f})."
        local_logger.warning(msg)

    return hist


def _safe_divide(a: np.ndarray, b: np.ndarray, ztol: float = 1e-8) -> np.ndarray:
    """Safe division to avoid division by zero"""

    return np.where(b > ztol, np.exp(np.log(a) - np.log(b)), 0.0)


@dataclass
class ImportanceSamplingData:
    """Importance sampling data"""

    posterior_samples: pd.DataFrame
    prior_samples: pd.DataFrame
    new_prior_samples: pd.DataFrame
    binsize_spin: float = 0.05
    binsize_mass: float = 1.0

    def get_binsize(self, col_name: str) -> Optional[float]:

        if col_name.startswith("a"):
            return self.binsize_spin

        if col_name.startswith("m"):
            return self.binsize_mass

    def get_nbins(self, col_name: str) -> Optional[int]:
        """Get the number of bins for a given column name"""

        binsize = self.get_binsize(col_name)

        if binsize is None:
            return None

        bounds = self.bounds[col_name]
        return int((bounds.high - bounds.low) / binsize)

    def _get_hist(self, samples: pd.Series) -> np.ndarray:
        """Get the histogram for a given column name"""

        nbins = self.get_nbins(samples.name)

        if nbins is None:
            raise ValueError(f"Unknown binsize for column {samples.name}")

        return get_histogram(samples, nbins=nbins, bounds=self.bounds[samples.name])

    def _compute_bounds(self) -> dict[str, Domain]:
        """Get the bounds of the dataframes"""

        bounds: dict[str, Domain] = {}

        for col in self.posterior_samples.columns:
            if col not in self.prior_samples.columns:
                prior_min = np.inf
                prior_max = -np.inf
            else:
                prior_min = self.prior_samples[col].min()
                prior_max = self.prior_samples[col].max()

            if col not in self.new_prior_samples.columns:
                new_prior_min = np.inf
                new_prior_max = -np.inf
            else:
                new_prior_min = self.new_prior_samples[col].min()
                new_prior_max = self.new_prior_samples[col].max()

            _min = min(self.posterior_samples[col].min(), prior_min, new_prior_min)
            _max = max(self.posterior_samples[col].max(), prior_max, new_prior_max)

            bounds[col] = Domain(low=_min, high=_max)

        return bounds

    def __post_init__(self) -> None:
        """Compute the bounds of the dataframes"""

        self._bounds = self._compute_bounds()

    @property
    def bounds(self) -> dict[str, Domain]:
        """Get the bounds of the dataframes"""

        return self._bounds

    def get_edges(self, col_name: str) -> np.ndarray:
        """Get the edges of the histogram for a given column name"""

        nbins = self.get_nbins(col_name)

        if nbins is None:
            raise ValueError(f"Unknown binsize for column {col_name}")

        bounds = self.bounds[col_name]
        return np.linspace(bounds.low, bounds.high, nbins + 1)

    def get_binwidth(self, col_name: str) -> float:
        """Get the bin width for a given column name"""

        edges = self.get_edges(col_name)
        return edges[1] - edges[0]

    @pre_release
    def get_likelihood_samples(self, random_state=42, ztol=1e-8) -> np.ndarray:
        """Get samples for likelihood function"""

        weights = np.ones(len(self.posterior_samples))

        for col in self.posterior_samples.columns:
            # Assume uniform prior if the column is not in prior samples
            if col not in self.prior_samples.columns:
                continue

            prior_hist = self._get_hist(self.prior_samples[col])
            rv = rv_histogram((_safe_divide(1.0, prior_hist, ztol=ztol), self.get_edges(col_name=col)))
            weights *= rv.pdf(self.posterior_samples[col])

        return self.posterior_samples.sample(
            n=len(self.posterior_samples),
            weights=weights,
            replace=True,
            random_state=random_state,
        )

    @pre_release
    def get_weights(self, col_name: str, ztol=1e-8, new_prior_samples: Optional[pd.Series] = None) -> np.ndarray:
        """Get the weights for the importance sampling"""

        weights = np.ones(len(self.posterior_samples))

        if (col_name not in self.prior_samples.columns) and (col_name not in self.new_prior_samples.columns):
            return weights

        prior_hist = self._get_hist(self.prior_samples[col_name]) if col_name in self.new_prior_samples.columns else 1.0
        if new_prior_samples:
            new_prior_hist = self._get_hist(new_prior_samples)
        else:
            new_prior_hist = (
                self._get_hist(self.new_prior_samples[col_name]) if col_name in self.new_prior_samples.columns else 1.0
            )
        # Avoid division by zero
        ratio = _safe_divide(new_prior_hist, prior_hist, ztol=ztol)
        rv = rv_histogram((ratio, self.get_edges(col_name)))
        weights *= rv.pdf(self.posterior_samples[col_name])

        return weights

    @pre_release
    def get_bayes_factor(self, ztol=1e-8, random_state=42) -> float:
        """Compute the Bayes factor between two models

        NOTE: In this implementation, the likelihood function remains untouched.
        So that the Bayes factor is computed as the ratio of the new prior to the old prior.
        Details please check importance sampling.
        """

        bf = 1.0
        weights = np.ones(len(self.prior_samples))

        for col in self.posterior_samples.columns:
            if col not in self.new_prior_samples.columns:
                continue

            prior_hist = self._get_hist(self.prior_samples[col])
            posterior_hist = self._get_hist(self.posterior_samples[col])
            new_prior_hist = self._get_hist(self.new_prior_samples[col])

            ratios = _safe_divide(new_prior_hist, prior_hist, ztol=ztol)
            rv = rv_histogram((ratios, self.get_edges(col_name=col)))
            weights *= rv.pdf(self.prior_samples[col])

            bf *= np.sum(new_prior_hist * _safe_divide(posterior_hist, prior_hist, ztol=ztol)) * self.get_binwidth(col)

        # Correction
        inferred_new_prior_samples = self.prior_samples.sample(
            n=len(self.prior_samples),
            weights=weights,
            replace=True,
            random_state=random_state,
        )
        for col in self.posterior_samples:
            if col in self.new_prior_samples.columns:
                continue

            prior_hist = self._get_hist(self.prior_samples[col])
            posterior_hist = self._get_hist(self.posterior_samples[col])
            new_prior_hist = self._get_hist(inferred_new_prior_samples[col])
            bf *= np.sum(new_prior_hist * _safe_divide(posterior_hist, prior_hist, ztol=ztol)) * self.get_binwidth(col)

        return bf

    @pre_release
    def get_reweighted_samples(self, ztol=1e-8, random_state=42) -> pd.DataFrame:
        """Get the reweighted samples for the importance sampling"""

        weights = np.ones(len(self.posterior_samples))

        for col in self.posterior_samples:
            weights *= self.get_weights(col_name=col, ztol=ztol)

        reweighted_samples = self.posterior_samples.sample(
            n=len(self.posterior_samples),
            weights=weights,
            replace=True,
            random_state=random_state,
        )

        return reweighted_samples


# Alias
ISData = ImportanceSamplingData



from typing import Optional

import corner
import matplotlib.colors as mcolors
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import archeo.logger
from archeo.constants import Columns as C
from archeo.constants import EscapeVelocity
from archeo.constants import Suffixes as S
from archeo.schema import Labels, Padding
from archeo.visualization import base


local_logger = archeo.logger.get_logger(__name__)


def filter_unmapped_samples(df: pd.DataFrame) -> pd.DataFrame:
    """Filter out the samples that are not mapped to the posterior.

    Args:
        df (pd.DataFrame): The prior dataframe.

    Returns:
        df (pd.DataFrame): The filtered prior dataframe.
    """

    return df.dropna(subset=[S.FINAL(C.KICK)])


def mass_estimates(
    df: pd.DataFrame,
    label: str,
    filename="mass_estimates",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the posterior mass estimates.

    Args:
        df (pd.DataFrame): The posterior dataframe.
        label (str): Label of the posterior.
        filename (str): Output filename.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    df = filter_unmapped_samples(df)

    padding = Padding(lpad=0.13, bpad=0.14)
    labels = Labels(
        title="Distribution of Estimated Masses",
        xlabel=r"Mass [$M_{\odot}$]",
        ylabel="PDF",
    )
    fig, ax = base.initialize_plot(figsize=(9, 4), labels=labels, padding=padding)
    colors = iter(mcolors.TABLEAU_COLORS.keys())

    col_to_name = {
        S.FINAL(C.MASS): label + ": ",
        S.PRIMARY(C.MASS): "Heavier Parent: ",
        S.SECONDARY(C.MASS): "Ligher Parent: ",
    }
    for col, name in col_to_name.items():
        _plot_pdf(ax, next(colors), df[col], name, unit=r"[$M_{\odot}$]")
    _add_pisn_gap(ax, next(colors))

    plt.legend()
    base.clear_default_labels(ax)
    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, ax)


def _add_pisn_gap(ax, color: str) -> None:
    """Add PISN gap to the plot.

    Args:
        ax (plt.Axes): Axes.
        color (str): Color.
    """

    ax.axvline(65.0, color=color, linewidth=0.9, linestyle="--", label="PISN Gap")
    ax.axvline(130.0, color=color, linewidth=0.9, linestyle="--")


def corner_estimates(  # pylint: disable=dangerous-default-value
    dfs: dict[str, pd.DataFrame],
    levels: list[float] = [0.68, 0.9],
    nbins: int = 70,
    filename="corner_estimates",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the posterior corner plot.

    Args:
        dfs (dict[str, pd.DataFrame]): Key: name of the posterior, value: posterior dataframe.
        levels (list[float]): Contour levels.
        nbins (int): Number of bins.
        filename (str): Output filename.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    corner_type_to_var_names = {
        "part": [S.PRIMARY(C.MASS), S.SECONDARY(C.MASS), S.FINAL(C.KICK)],
        "full": [
            S.PRIMARY(C.MASS),
            S.SECONDARY(C.MASS),
            S.FINAL(C.MASS),
            S.FINAL(C.KICK),
            S.FINAL(C.SPIN_MAG),
            S.EFF(C.SPIN),
        ],
    }
    corner_type_to_labels = {
        "part": [
            r"$m_1$ [$M_{\odot}$]",
            r"$m_2$ [$M_{\odot}$]",
            r"$v_f$ [km s$^{-1}$]",
        ],
        "full": [
            r"$m_1$ [$M_{\odot}$]",
            r"$m_2$ [$M_{\odot}$]",
            r"$m_f$ [$M_{\odot}$]",
            r"$v_f$ [km s$^{-1}$]",
            "$a_f$",
            "$\\chi_{eff}$",
        ],
    }

    # Add precession spin if available
    prec_spins = [df[S.PREC(C.SPIN)].max() for df in dfs.values()]
    if max(prec_spins) > 0.0:
        if np.isclose(min(prec_spins), 0.0):
            local_logger.warning(
                "Precession spin is not available for all dataframes, " "skipped adding it to corner plot."
            )
        else:
            corner_type_to_var_names["full"].append(S.PREC(C.SPIN))
            corner_type_to_labels["full"].append("$\\chi_{p}$")

    for corner_type, var_names in corner_type_to_var_names.items():
        fig, axes = base.initialize_plot(ncols=len(var_names), nrows=len(var_names), figsize=(9, 9))
        colors = iter(mcolors.TABLEAU_COLORS.keys())
        handles = []

        for label, df in dfs.items():
            df = filter_unmapped_samples(df)

            if len(df) < nbins:
                local_logger.info("Dataframe does not have enough samples to plot.")
                continue

            color = next(colors)
            corner.corner(
                data=df[var_names],
                bins=nbins,
                var_names=var_names,
                labels=corner_type_to_labels[corner_type],
                levels=levels,
                plot_density=True,
                plot_samples=False,
                color=color,
                fill_contours=False,
                smooth=True,
                plot_datapoints=False,
                hist_kwargs=dict(density=True),
                fig=fig,
            )
            handles.append(mlines.Line2D([], [], color=color, label=label))

        fig.legend(
            handles=handles,
            fontsize=15,
            frameon=True,
            bbox_to_anchor=(1.0, 0.95),
            loc="upper right",
        )
        base.savefig_and_close(f"{corner_type}_{filename}", output_dir, close, fmt)

    return (fig, axes)


def second_generation_probability_curve(
    dfs: dict[str, pd.DataFrame],
    filename="second_generation_probability_curve",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the second generation probability curve.

    Args:
        dfs (list[pd.DataFrame]): Key: name of the posterior, value: posterior dataframe.
        label (list[str]): Label of each posterior.
        filename (str): Output filename.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    # Set up x-axis
    x_max = max(df[S.FINAL(C.KICK)].max() for df in dfs.values())
    x = np.linspace(0.0, x_max, 300)

    padding = Padding(bpad=0.14)
    labels = Labels(
        title="Second Generation Probability Curve",
        xlabel="Escape Velocity $v_{esc}$ [km s$^{-1}$]",
        ylabel="Second Generation Probability $p_{2g}$",
    )
    fig, ax = base.initialize_plot(figsize=(10, 8), labels=labels, padding=padding, fontsize=15)
    colors = iter(mcolors.TABLEAU_COLORS.keys())

    for label, df in dfs.items():
        df = filter_unmapped_samples(df)

        # Calculate the CDF
        y = []
        for kick in x:
            df_samples = df.loc[
                (df[S.FINAL(C.KICK)] <= kick) & (df[S.PRIMARY(C.MASS)] <= 65) & (df[S.SECONDARY(C.MASS)] <= 65)
            ]
            if df_samples.empty:
                y.append(0.0)
            else:
                y.append(len(df_samples) / len(df))

        # Plot the CDF
        sns.lineplot(y=y, x=x, ax=ax, color=next(colors), label=label)

    base.add_escape_velocity(ax, x_max, max(y))

    plt.legend()
    base.clear_default_labels(ax)
    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, ax)


def effective_spin_estimates(
    dfs: dict[str, pd.DataFrame],
    filename="effective_spin_estimates",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the effective spin PDF.

    Args:
        dfs (dict[str, pd.DataFrame]): Key: name of the posterior, value: posterior dataframe.
        filename (str): Output filename.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    padding = Padding(bpad=0.14)
    labels = Labels(
        title="Effective Spin PDF",
        xlabel="Effective Spin $\\chi_{eff}$",
        ylabel="PDF",
    )
    fig, ax = base.initialize_plot(figsize=(10, 8), labels=labels, padding=padding, fontsize=15)
    colors = iter(mcolors.TABLEAU_COLORS.keys())

    for label, df in dfs.items():
        _plot_pdf(ax, next(colors), df[S.EFF(C.SPIN)], label)

    plt.legend()
    base.clear_default_labels(ax)
    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, ax)


def precession_spin_estimates(
    dfs: dict[str, pd.DataFrame],
    filename="precession_spin_estimates",
    output_dir: Optional[str] = None,
    close: bool = True,
    fmt: str = "png",
):
    """Plot the precession spin PDF.

    Args:
        dfs (dict[str, pd.DataFrame]): Key: name of the posterior, value: posterior dataframe.
        filename (str): Output filename.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "png".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    padding = Padding(bpad=0.14)
    labels = Labels(
        title="Precession Spin PDF",
        xlabel="Precession Spin $\\chi_p$",
        ylabel="PDF",
    )
    fig, ax = base.initialize_plot(figsize=(10, 8), labels=labels, padding=padding, fontsize=15)
    colors = iter(mcolors.TABLEAU_COLORS.keys())

    for label, df in dfs.items():
        _plot_pdf(ax, next(colors), df[S.PREC(C.SPIN)], label)

    plt.legend()
    base.clear_default_labels(ax)
    base.savefig_and_close(filename, output_dir, close, fmt)
    return (fig, ax)


def _plot_pdf(
    ax,
    color: str,
    series: pd.Series,
    name: str,
    unit: Optional[str] = None,
):
    """Plot the PDF of a parameter.

    Args:
        ax (plt.Axes): Axes.
        color (str): Color.
        series (pd.Series): Series (pdf).
        name (str): Name.
        unit (Optional[str]): Unit.
    """

    _series = series.dropna()
    density, bins = np.histogram(a=_series, bins=70, density=True)
    low, mid, high = (
        _series.quantile(0.05),
        _series.quantile(0.5),
        _series.quantile(0.95),
    )
    label = "%s: $%.2f_{-%.2f}^{+%.2f}$" % (name, mid, mid - low, high - mid)
    if unit:
        label += f" {unit}"
    ax.stairs(density, bins, label=label, color=color)


def table_estimates(
    dfs: dict[str, pd.DataFrame],
    filename="table_estimates",
    output_dir: Optional[str] = None,
    fmt: str = "md",
) -> pd.DataFrame:
    """Plot the posterior mass estimates.

    Args:
        dfs (dict[str, pd.DataFrame]): Key: name of the posterior, value: posterior dataframe.
        filename (str): Output filename.
        output_dir (Optional[str]): Output directory.
        close (bool): Whether to close the figure.
        fmt (str): The format of the visualizations. Defaults to "md".

    Returns:
        fig (plt.Figure): Figure.
        axes (plt.Axes): Axes.
    """

    col_to_names = {
        S.PRIMARY(C.MASS): "$m_1$",
        S.SECONDARY(C.MASS): "$m_2$",
        C.MASS_RATIO: "$q$",
        S.FINAL(C.MASS): "$m_f$",
        S.FINAL(C.SPIN_MAG): "$a_f$",
        S.FINAL(C.KICK): "$v_f$",
        S.PREC(C.SPIN): "$\\chi_{p}$",
        S.EFF(C.SPIN): "$\\chi_{eff}$",
    }
    data = {
        "": dfs.keys(),
        "Recovery Rate": [df[C.RECOVERY_RATE].iloc[0] for df in dfs.values()],
        **{f"p2g_{v_esc.short()}": [v_esc.compute_p2g(df) for df in dfs.values()] for v_esc in EscapeVelocity},
    }

    for col, name in col_to_names.items():
        data[name] = []
        for df in dfs.values():
            df = filter_unmapped_samples(df)
            low, mid, high = (
                df[col].quantile(0.05),
                df[col].quantile(0.5),
                df[col].quantile(0.95),
            )
            value = "$%.2f_{-%.2f}^{+%.2f}$" % (mid, mid - low, high - mid)
            data[name].append(value)

    df_table = pd.DataFrame(data)

    if output_dir:
        if fmt.lower() == "md":
            df_table.to_markdown(f"{output_dir}/{filename}.{fmt}", index=False)

        elif fmt.lower() == "csv":
            df_table.to_csv(f"{output_dir}/{filename}.{fmt}", index=False)

        else:
            local_logger.warning("Unsupported format: %s.", fmt)

    return df_table



from typing import Callable

import numpy as np
import pandas as pd

import archeo.logger
from archeo.schema import Domain


local_logger = archeo.logger.get_logger(__name__)


def get_mahapatra_mass_fn(mass: Domain, alpha=2.3, dm=4.83, n_samples=500000) -> Callable:
    """Get a mass function from Mahapatra's mass distribution.
    NOTE: For details, see https://arxiv.org/abs/2209.05766.

    Args:
        mass (Domain): Mass domain.
        alpha (float): Power law index.
        dm (float): Tapering parameter.
        n_samples (int): Number of samples to generate.

    Returns:
        mass_fn (Callable): Mass function.
    """

    def _f(ds: pd.Series) -> pd.Series:
        """Calculate the function f in Mahapatra's paper"""

        mp = ds - mass.low
        return np.exp(dm / mp + dm / (mp - dm))

    def smoothing_func(ds: pd.Series) -> pd.Series:
        """Smoothing function."""

        probis = ds.copy()
        probis[ds < mass.low + dm] = 1 / (_f(ds[ds < mass.low + dm]) + 1)  # type: ignore
        probis[ds > mass.low + dm] = 1  # type: ignore
        probis *= ds ** (-alpha)
        return probis

    masses = pd.Series(np.random.uniform(mass.low, mass.high, size=n_samples))
    probis = smoothing_func(masses)
    probis /= probis.sum()

    def mass_from_mahapatra() -> float:
        """Generate a mass from Mahapatra's mass distribution."""

        return np.random.choice(masses, p=probis)

    return mass_from_mahapatra



import logging
from dataclasses import asdict
from typing import Callable, Optional

import numpy as np
import pandas as pd
from tqdm import tqdm

from archeo.constants import Columns as C
from archeo.constants import Suffixes as S
from archeo.core.mahapatra import get_mahapatra_mass_fn
from archeo.schema import Binary, Event, PriorConfig
from archeo.utils.helper import pre_release
from archeo.utils.math import sph2cart
from archeo.utils.parallel import multithread_run


logger = logging.getLogger(__name__)


class Simulator:
    """Simulator class to simulate binary black hole mergers"""

    def __init__(self, prior_config: PriorConfig):
        """Initializes the simulator with the given prior configuration

        Args:
            prior_config (PriorConfig): The prior configuration to use
        """

        self._prior_config = prior_config

        self._fits = self._prior_config.fits.load()
        self._n_samples = self._prior_config.n_samples

        if self._prior_config.is_mahapatra:
            self._m1_fn = get_mahapatra_mass_fn(mass=self._prior_config.m_1)
            self._m2_fn = get_mahapatra_mass_fn(mass=self._prior_config.m_2)
        else:
            self._m1_fn = self._prior_config.m_1.draw
            self._m2_fn = self._prior_config.m_2.draw

        # Dummy functions
        self._is_remnant_1 = False
        self._r1_fn = lambda: (0, 0, 0)
        self._is_remnant_2 = False
        self._r2_fn = lambda: (0, 0, 0)

        self._chi1_fns = {
            "magnitude": self._prior_config.a_1.draw,
            "theta": self._prior_config.theta_1.draw,
            "phi": self._prior_config.phi_1.draw,
        }
        self._chi2_fns = {
            "magnitude": self._prior_config.a_2.draw,
            "theta": self._prior_config.theta_2.draw,
            "phi": self._prior_config.phi_2.draw,
        }

        self._m2_bounds = self._prior_config.m_2
        self._q_bounds = self._prior_config.mass_ratio

        self._is_spin_aligned = self._prior_config.is_spin_aligned
        self._only_up_aligned_spin = self._prior_config.is_only_up_aligned_spin
        self._is_uniform_in_q = self._prior_config.is_uniform_in_mass_ratio

    def __call__(self) -> Event:
        """Simulates a binary black hole merger event

        Returns:
            Event: The simulated event
        """

        b = self._get_binary()

        q = b.m_1 / b.m_2
        v_f, v_f_err = self._fits.vf(q, b.chi_1, b.chi_2)
        chi_f, chi_f_err = self._fits.chif(q, b.chi_1, b.chi_2)
        m_ret, m_ret_err = self._fits.mf(q, b.chi_1, b.chi_2)

        return Event(
            m_1=b.m_1,
            m_2=b.m_2,
            k_1=b.k_1,
            k_2=b.k_2,
            m_ret=m_ret,
            m_ret_err=m_ret_err,
            v_f=v_f,
            v_f_err=v_f_err,
            chi_1=b.chi_1,
            chi_2=b.chi_2,
            chi_f=chi_f,
            chi_f_err=chi_f_err,
        )

    def simulate(self, use_threads=True, n_threads: Optional[int] = None) -> pd.DataFrame:
        """Simulates multiple binary black hole merger events"""

        if use_threads:
            events = multithread_run(self, [{} for _ in range(self._n_samples)], n_threads)
        else:
            events = [self() for _ in tqdm(range(self._n_samples))]

        df = pd.DataFrame([asdict(event) for event in events])
        return df

    def _get_binary(self) -> Binary:
        """Draws a binary from the prior distribution

        Returns:
            Binary: The drawn binary
        """

        if self._is_uniform_in_q:
            m_1, a_1, k_1, m_2, a_2, k_2 = self._get_params_unif_q()
        else:
            m_1, a_1, k_1, m_2, a_2, k_2 = self._get_params_non_unif_q()

        chi_uv_1, chi_uv_2 = self._get_spin_uv(self._chi1_fns), self._get_spin_uv(self._chi2_fns)
        chi_1 = (chi_uv_1[0] * a_1, chi_uv_1[1] * a_1, chi_uv_1[2] * a_1)
        chi_2 = (chi_uv_2[0] * a_2, chi_uv_2[1] * a_2, chi_uv_2[2] * a_2)

        return Binary(m_1=m_1, m_2=m_2, chi_1=chi_1, chi_2=chi_2, k_1=k_1, k_2=k_2)

    def _get_spin_uv(self, fns: dict[str, Callable]) -> tuple[float, float, float]:
        """Draws the spin of the binary (unit vector)

        Args:
            fns (dict[str, Callable]): The functions to draw the spin

        Returns:
            tuple[float, float, float]: The drawn spin unit vector
        """

        if self._is_spin_aligned:
            if self._only_up_aligned_spin:
                return (0, 0, 1)

            direction = np.random.choice([-1, 1])
            return (0, 0, direction)

        theta = np.arccos(-1 + 2 * fns["theta"]())
        phi = fns["phi"]() * np.pi
        univ = sph2cart(theta=theta, phi=phi)
        return tuple(univ)

    def _get_params_non_unif_q(self) -> tuple[float, float, float, float, float, float]:
        """Draws the masses of the binary from the mass functions

        Returns:
            tuple[float, float]: The drawn masses
        """

        if not self._is_remnant_1 and not self._is_remnant_2:
            m_1, m_2 = (self._m1_fn(), self._m2_fn())
            a_1, a_2 = self._chi1_fns["magnitude"](), self._chi2_fns["magnitude"]()
            k_1, k_2 = (0, 0)
        elif self._is_remnant_1 and self._is_remnant_2:
            m_1, a_1, k_1 = self._r1_fn()
            m_2, a_2, k_2 = self._r2_fn()
        elif self._is_remnant_1:
            m_1, a_1, k_1 = self._r1_fn()
            m_2, a_2, k_2 = self._m2_fn(), self._chi2_fns["magnitude"](), 0.0
        # Only self._is_remnant_2
        else:
            m_1, a_1, k_1 = self._m1_fn(), self._chi1_fns["magnitude"](), 0.0
            m_2, a_2, k_2 = self._r2_fn()

        # If not, resample the masses (recursion)
        if self._prior_config.is_masses_swappable:
            m_1, m_2 = max(m_1, m_2), min(m_1, m_2)

        q = m_1 / m_2
        # Check whether the mass ratio is in the domain
        if not self._q_bounds.contain(q):
            return self._get_params_non_unif_q()

        return (m_1, a_1, k_1, m_2, a_2, k_2)

    def _get_params_unif_q(self) -> tuple[float, float, float, float, float, float]:
        """Draws the masses of the binary from the mass ratio function

        Returns:
            tuple[float, float]: The drawn masses
        """

        q = self._q_bounds.draw()
        m_1 = self._m1_fn()
        m_2 = m_1 / q

        # Check whether the masses are in the domain
        # If not, resample the masses (recursion)
        if not self._m2_bounds.contain(m_2):
            return self._get_params_unif_q()

        a_1 = self._chi1_fns["magnitude"]()
        a_2 = self._chi2_fns["magnitude"]()
        k_1, k_2 = (0.0, 0.0)

        return (m_1, a_1, k_1, m_2, a_2, k_2)

    @staticmethod
    @pre_release
    def get_draw_func(df: pd.DataFrame) -> Callable[[], tuple[float, float, float]]:
        """Returns a function to draw remnant properties from the dataframe"""

        def draw() -> tuple[float, float, float]:
            """Draws the remnant mass and spin"""

            idx = np.random.random_integers(low=0, high=len(df) - 1)
            return (
                df.loc[idx, S.FINAL(C.MASS)],
                df.loc[idx, S.FINAL(C.SPIN_MAG)],
                df.loc[idx, S.FINAL(C.KICK)],
            )

        return draw

    @pre_release
    def use_remnant_results(
        self,
        df_bh1: Optional[pd.DataFrame] = None,
        df_bh2: Optional[pd.DataFrame] = None,
        kick_limit: Optional[float] = None,
    ) -> None:
        """Uses the remnant results from the given file

        Args:
            df (Optional[pd.DataFrame]): The dataframe containing the remnant results
            filepath (Optional[str]): The path to the file containing the remnant results
            bh (Literal[1, 2]): The black hole to use the results for
            kick_limit (Optional[float]): The kick limit to apply
        """

        if self._is_uniform_in_q:
            raise ValueError("Cannot use remnant results with uniform mass ratio prior")

        if df_bh1 is not None:
            df_bh1 = df_bh1[df_bh1[S.FINAL(C.KICK)] <= (kick_limit or np.inf)][
                [S.FINAL(C.SPIN_MAG), S.FINAL(C.MASS), S.FINAL(C.KICK)]
            ].reset_index(drop=True)
            self._is_remnant_1 = True
            self._r1_fn = self.get_draw_func(df_bh1)
            logger.warning(
                "Using remnant results for black hole 1. "
                "This will override the prior configuration for black hole 1."
            )

        if df_bh2 is not None:
            df_bh2 = df_bh2[df_bh2[S.FINAL(C.KICK)] <= (kick_limit or np.inf)][
                [S.FINAL(C.SPIN_MAG), S.FINAL(C.MASS), S.FINAL(C.KICK)]
            ].reset_index(drop=True)
            self._is_remnant_2 = True
            self._r2_fn = self.get_draw_func(df_bh2)
            logger.warning(
                "Using remnant results for black hole 2. "
                "This will override the prior configuration for black hole 2."
            )

    @property
    def model_name(self) -> str:
        """Returns the name of the model used for simulation"""

        return self._prior_config.fits.value



from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy.stats import ks_2samp

import archeo.logger
from archeo.constants import SPEED_OF_LIGHT
from archeo.constants import Columns as C
from archeo.constants import Prefixes as P
from archeo.constants import Suffixes as S
from archeo.core.simulator import Simulator
from archeo.preset import get_prior_config
from archeo.schema import PriorConfig
from archeo.utils.parallel import multithread_run


local_logger = archeo.logger.get_logger(__name__)


class Prior(pd.DataFrame):
    """A class to represent the prior distribution."""

    def __init__(
        self,
        *args,
        rescale_mass: bool = False,
        sample_ratio: int = 1,
        spin_tolerance: float = 0.05,  # unit: dimensionless
        mass_tolerance: float = 1.0,  # unit: solar mass
        **kwargs,
    ) -> None:
        """Construct a prior dataframe.

        Args:
            rescale_mass (bool): Whether to enable mass marginalization
            sample_ratio (int): The number of samples to be sampled each time
            spin_tolerance (float): The tolerance of the spin
            mass_tolerance (float): The tolerance of the mass
        """

        super().__init__(*args, **kwargs)

        self._rescale_mass = rescale_mass
        self._sample_ratio = sample_ratio
        self._spin_tolerance = spin_tolerance
        self._mass_tolerance = mass_tolerance

    def _sample_from_possible_samples(self, df: pd.DataFrame) -> pd.DataFrame:
        """Sample from a dataframe.

        Args:
            df (pd.DataFrame): The dataframe to sample from.

        Returns:
            df (pd.DataFrame): The sampled dataframe.
        """

        if df.empty:
            local_logger.warning("No similar samples in the prior.")
            # Return a number of samples with nan values
            df = pd.DataFrame(index=range(self._sample_ratio), columns=df.columns)
        else:
            df = df.sample(self._sample_ratio, replace=True)

        return df

    def retrieve_samples(self, spin_measure: float, mass_measure: float) -> pd.DataFrame:
        """Retrieve the samples from prior.

        Args:
            spin_measure (float): The measured spin
            mass_measure (float): The measured mass

        Returns:
            pd.DataFrame: The sampled dataframe
        """

        if not self._rescale_mass:
            # Find the possible samples in the prior
            # Based on:
            #    1. mass_prior - tol < mass_measure < mass_prior + tol
            #    2. spin_prior - tol < spin_measure < spin_prior + tol
            possible_samples = self.loc[
                ((self[S.FINAL(C.MASS)] - mass_measure).abs() < self._mass_tolerance)
                & ((self[S.FINAL(C.SPIN_MAG)] - spin_measure).abs() < self._spin_tolerance)
            ]
            likelihood = len(possible_samples) / len(self)

            # Sample n_sample samples from the possible samples
            samples = self._sample_from_possible_samples(possible_samples)

            samples[C.LIKELIHOOD] = likelihood
            samples[P.ORIGINAL(S.FINAL(C.SPIN_MAG))] = spin_measure
            samples[P.ORIGINAL(S.FINAL(C.MASS))] = mass_measure

            return samples

        # Find the possible samples in the prior
        # Based on:
        #    1. spin_prior - tol < spin_measure < spin_prior + tol
        possible_samples = self.loc[(self[S.FINAL(C.SPIN_MAG)] - spin_measure).abs() < self._spin_tolerance]
        likelihood = len(possible_samples) / len(self)

        # Sample n_sample samples from the possible samples
        samples = self._sample_from_possible_samples(possible_samples)

        # Calculate the mass parameters (for mass not injected case)
        samples[S.PRIMARY(C.MASS)] = (
            mass_measure / samples[S.RETAINED(C.MASS)] * samples[C.MASS_RATIO] / (1 + samples[C.MASS_RATIO])
        )
        samples[S.SECONDARY(C.MASS)] = mass_measure / samples[S.RETAINED(C.MASS)] / (1 + samples[C.MASS_RATIO])
        samples[S.FINAL(C.MASS)] = mass_measure

        samples[C.LIKELIHOOD] = likelihood
        samples[P.ORIGINAL(S.FINAL(C.SPIN_MAG))] = spin_measure
        samples[P.ORIGINAL(S.FINAL(C.MASS))] = mass_measure

        return samples

    @property
    def _constructor(self):
        """Return the constructor of the class."""

        return pd.DataFrame

    @classmethod
    def from_feather(cls, path: str, **kwargs) -> "Prior":
        """Read the feather file.

        Args:
            path (str): The path to the feather file.
            **kwargs: The keyword arguments for parental class (pd.DataFrame).

        Returns:
            Prior: The prior distribution.
        """

        return cls(pd.read_feather(path), **kwargs)

    @classmethod
    def from_csv(cls, path: str, **kwargs) -> "Prior":
        """Read the csv file.

        Args:
            path (str): The path to the csv file.
            **kwargs: The keyword arguments for parental class (pd.DataFrame).

        Returns:
            Prior: The prior distribution
        """

        return cls(pd.read_csv(path), **kwargs)

    @classmethod
    def from_json(cls, path: str, **kwargs) -> "Prior":
        """Read the json file.

        Args:
            path (str): The path to the json file.
            **kwargs: The keyword arguments for parental class (pd.DataFrame).

        Returns:
            Prior: The prior distribution.
        """

        return cls(pd.read_json(path), **kwargs)

    @classmethod
    def from_parquet(cls, path: str, **kwargs) -> "Prior":
        """Read the parquet file.

        Args:
            path (str): The path to the parquet file.
            **kwargs: The keyword arguments for parental class (pd.DataFrame).

        Returns:
            Prior: The prior distribution.
        """

        return cls(pd.read_parquet(path), **kwargs)

    @classmethod
    def from_filepath(cls, path: str, **kwargs) -> "Prior":
        """Read the file.

        Args:
            path (str): The path to the file.
            **kwargs: The keyword arguments for parental class (pd.DataFrame).

        Returns:
            Prior: The prior distribution.
        """

        if path.endswith(".feather"):
            return cls.from_feather(path, **kwargs)

        if path.endswith(".csv"):
            return cls.from_csv(path, **kwargs)

        if path.endswith(".json"):
            return cls.from_json(path, **kwargs)

        if path.endswith(".parquet"):
            return cls.from_parquet(path, **kwargs)

        raise ValueError(f"Unsupported file format: {path}")

    @classmethod
    def from_config(cls, prior_config: Union[PriorConfig, str], use_threads=True, **kwargs) -> "Prior":
        """Generate the prior from the prior config.

        Args:
            prior_config (PriorConfig): The prior configuration.
            use_threads (bool): Whether to use threads.
            **kwargs: The keyword arguments for the class.

        Returns:
            Prior: The prior distribution.
        """

        if isinstance(prior_config, str):
            prior_config = get_prior_config(prior_config)

        simulator = Simulator(prior_config)
        return cls.from_simulator(simulator, use_threads=use_threads, **kwargs)

    @classmethod
    def from_simulator(cls, simulator: Simulator, use_threads=True, **kwargs) -> "Prior":
        """Generate the prior from the simulator.

        Args:
            simulator (Simulator): The simulator.
            use_threads (bool): Whether to use threads.
            **kwargs: The keyword arguments for the class.

        Returns:
            Prior: The prior distribution.
        """

        df = cls(simulator.simulate(use_threads=use_threads), **kwargs)

        # Extract more information from the samples

        # Define dummy columns
        df[C.RECOVERY_RATE] = float("nan")
        df[C.KS_TEST_FOR_MASS] = float("nan")
        df[C.KS_PV_FOR_MASS] = float("nan")
        df[C.KS_TEST_FOR_SPIN] = float("nan")
        df[C.KS_PV_FOR_SPIN] = float("nan")
        df[C.SAMPLE_ID] = float("nan")

        # Store model name
        df[C.MODEL_NAME] = simulator.model_name

        # Calculate the mass ratio
        m1, m2 = df[S.PRIMARY(C.MASS)], df[S.SECONDARY(C.MASS)]
        df[C.MASS_RATIO] = q = m1 / m2

        # Calculate the remnant mass
        df[S.FINAL(C.MASS)] = df[S.RETAINED(C.MASS)] * (m1 + m2)

        # Calculate the BH kick velocity
        df[S.FINAL(C.KICK)] = df[S.FINAL(C.VELOCITY)].apply(lambda vf: np.sqrt(np.dot(vf, vf)) * SPEED_OF_LIGHT)

        # Calculate the BH spin
        df[S.FINAL(C.SPIN_MAG)] = df[S.FINAL(C.SPIN)].apply(lambda vf: np.sqrt(np.dot(vf, vf)))

        # Calculate the parental spins
        df[S.PRIMARY(C.SPIN_MAG)] = df[S.PRIMARY(C.SPIN)].apply(lambda chi: np.sqrt(np.dot(chi, chi)))
        df[S.SECONDARY(C.SPIN_MAG)] = df[S.SECONDARY(C.SPIN)].apply(lambda chi: np.sqrt(np.dot(chi, chi)))

        # Calculate the effective spin
        a1z = df[S.PRIMARY(C.SPIN)].apply(lambda chi: chi[-1])
        a2z = df[S.SECONDARY(C.SPIN)].apply(lambda chi: chi[-1])
        df[S.EFF(C.SPIN)] = (m1 * a1z + m2 * a2z) / (m1 + m2)

        # Calculate the precession spin
        a1h = df[S.PRIMARY(C.SPIN)].apply(lambda chi: np.sqrt(chi[0] ** 2 + chi[1] ** 2))
        a2h = df[S.SECONDARY(C.SPIN)].apply(lambda chi: np.sqrt(chi[0] ** 2 + chi[1] ** 2))
        df[S.PREC(C.SPIN)] = np.maximum(a1h, (4 / q + 3) / (3 / q + 4) / q * a2h)

        # Calculate minimum escape velocity of host environment
        df[P.MIN(S.ESC(C.VELOCITY))] = df.apply(
            lambda x: max(
                x[S.FINAL(C.KICK)],
                x[S.PRIMARY(C.KICK)],
                x[S.SECONDARY(C.KICK)],
            ),
            axis=1,
        )

        return df

    def to_posterior(
        self,
        mass_posterior: list[float],
        spin_posterior: list[float],
        use_threads=True,
        n_threads: Optional[int] = None,
    ) -> pd.DataFrame:
        """Convert the prior to the posterior.

        Args:
            mass_posterior (list[float]): The posterior mass.
            spin_posterior (list[float]): The posterior spin.
            use_threads (bool): Whether to use threads.
            n_threads (Optional[int]): The number of threads to be used.

        Returns:
            pd.DataFrame: The posterior distribution.
        """

        if use_threads:
            input_kwargs = [
                dict(spin_measure=spin_measure, mass_measure=mass_measure)
                for spin_measure, mass_measure in zip(spin_posterior, mass_posterior)
            ]
            samples = multithread_run(
                func=self.retrieve_samples,
                input_kwargs=input_kwargs,
                n_threads=n_threads,
            )

        else:
            samples = [
                self.retrieve_samples(spin_measure=spin_measure, mass_measure=mass_measure)
                for spin_measure, mass_measure in zip(spin_posterior, mass_posterior)
            ]

        df_posterior = pd.concat(samples)

        df_posterior[C.RECOVERY_RATE] = df_posterior[S.FINAL(C.KICK)].notna().sum() / len(df_posterior)

        ks, p_value = ks_2samp(
            df_posterior[P.ORIGINAL(S.FINAL(C.SPIN_MAG))],
            df_posterior[S.FINAL(C.SPIN_MAG)],
        )
        df_posterior[C.KS_TEST_FOR_SPIN] = ks
        df_posterior[C.KS_PV_FOR_SPIN] = p_value

        ks, p_value = ks_2samp(df_posterior[P.ORIGINAL(S.FINAL(C.MASS))], df_posterior[S.FINAL(C.MASS)])
        df_posterior[C.KS_TEST_FOR_MASS] = ks
        df_posterior[C.KS_PV_FOR_MASS] = p_value

        df_posterior[C.SAMPLE_ID] = df_posterior.apply(
            lambda x: x.name if pd.notna(x[S.FINAL(C.KICK)]) else None, axis=1
        )
        df_posterior = df_posterior.reset_index(drop=True)

        return df_posterior
